@@ -18,6 +18,7 @@ details. */
 #include "pinfo.h"
 #include "shared_info.h"
 #include "tls_pbuf.h"
+#include "sigproc.h"
 #include <assert.h>
 
 /* This is only to be used for writing.  When reading,
@@ -247,9 +248,9 @@ fhandler_pipe::fadvise (off_t offset, off_t length, int advice)
 }
 
 int
-fhandler_pipe::ftruncate (off_t length, bool allow_truncate)
+fhandler_pipe::fallocate (int mode, off_t offset, off_t length)
 {
-  return allow_truncate ? EINVAL : ESPIPE;
+  return (mode & __FALLOC_FL_TRUNCATE) ? EINVAL : ESPIPE;
 }
 
 char *
@@ -587,6 +588,17 @@ fhandler_pipe::fixup_after_fork (HANDLE parent)
   ReleaseMutex (hdl_cnt_mtx);
 }
 
+void
+fhandler_pipe::fixup_after_exec ()
+{
+  /* Set read pipe itself always non-blocking for cygwin process.
+     Blocking/non-blocking is simulated in raw_read(). For write
+     pipe, follow is_nonblocking(). */
+  bool mode = get_device () == FH_PIPEW ? is_nonblocking () : true;
+  set_pipe_non_blocking (mode);
+  fhandler_base::fixup_after_exec ();
+}
+
 int
 fhandler_pipe::dup (fhandler_base *child, int flags)
 {
@@ -1190,66 +1202,31 @@ cache_err:
 	       &pipename_key, &pipename_pid, &pipename_id) != 3)
     return NULL; /* Non cygwin pipe? */
 
-  if (wincap.has_query_process_handle_info ())
-    return get_query_hdl_per_process (name, ntfn); /* Since Win8 */
-  else
-    return get_query_hdl_per_system (name, ntfn); /* Win7 */
+  return get_query_hdl_per_process (name, ntfn); /* Since Win8 */
 }
 
-/* This function is faster than get_query_hdl_per_system(), however,
-   only works since Windows 8 because ProcessHandleInformation is not
-   suppoted by NtQueryInformationProcess() before Windows 8. */
 HANDLE
 fhandler_pipe::get_query_hdl_per_process (WCHAR *name,
 					  OBJECT_NAME_INFORMATION *ntfn)
 {
-  NTSTATUS status;
-  ULONG len;
-  DWORD n_process = 256;
-  PSYSTEM_PROCESS_INFORMATION spi;
-  do
-    { /* Enumerate processes */
-      DWORD nbytes = n_process * sizeof (SYSTEM_PROCESS_INFORMATION);
-      spi = (PSYSTEM_PROCESS_INFORMATION) HeapAlloc (GetProcessHeap (),
-						     0, nbytes);
-      if (!spi)
-	return NULL;
-      status = NtQuerySystemInformation (SystemProcessInformation,
-					 spi, nbytes, &len);
-      if (NT_SUCCESS (status))
-	break;
-      HeapFree (GetProcessHeap (), 0, spi);
-      n_process *= 2;
-    }
-  while (n_process < (1L<<20) && status == STATUS_INFO_LENGTH_MISMATCH);
-  if (!NT_SUCCESS (status))
-    return NULL;
+  winpids pids ((DWORD) 0);
 
-  /* In most cases, it is faster to check the processes in reverse order.
-     To do this, store PIDs into an array. */
-  DWORD *proc_pids = (DWORD *) HeapAlloc (GetProcessHeap (), 0,
-					  n_process * sizeof (DWORD));
-  if (!proc_pids)
+  /* In most cases, it is faster to check the processes in reverse order. */
+  for (LONG i = (LONG) pids.npids - 1; i >= 0; i--)
     {
-      HeapFree (GetProcessHeap (), 0, spi);
-      return NULL;
-    }
-  PSYSTEM_PROCESS_INFORMATION p = spi;
-  n_process = 0;
-  while (true)
-    {
-      proc_pids[n_process++] = (DWORD)(intptr_t) p->UniqueProcessId;
-      if (!p->NextEntryOffset)
-	break;
-      p = (PSYSTEM_PROCESS_INFORMATION) ((char *) p + p->NextEntryOffset);
-    }
-  HeapFree (GetProcessHeap (), 0, spi);
+      NTSTATUS status;
+      ULONG len;
+
+      /* Non-cygwin app may call ReadFile() for empty pipe, which makes
+	NtQueryObject() for ObjectNameInformation block. Therefore, do
+	not try to get query_hdl for non-cygwin apps. */
+      _pinfo *p = pids[i];
+      if (!p || ISSTATE (p, PID_NOTCYGWIN))
+	continue;
 
-  for (LONG i = (LONG) n_process - 1; i >= 0; i--)
-    {
       HANDLE proc = OpenProcess (PROCESS_DUP_HANDLE
 				 | PROCESS_QUERY_INFORMATION,
-				 0, proc_pids[i]);
+				 0, p->dwProcessId);
       if (!proc)
 	continue;
 
@@ -1313,7 +1290,6 @@ fhandler_pipe::get_query_hdl_per_process (WCHAR *name,
 	      query_hdl_proc = proc;
 	      query_hdl_value = (HANDLE)(intptr_t) phi->Handles[j].HandleValue;
 	      HeapFree (GetProcessHeap (), 0, phi);
-	      HeapFree (GetProcessHeap (), 0, proc_pids);
 	      return h;
 	    }
 close_handle:
@@ -1323,77 +1299,56 @@ close_handle:
 close_proc:
       CloseHandle (proc);
     }
-  HeapFree (GetProcessHeap (), 0, proc_pids);
   return NULL;
 }
 
-/* This function is slower than get_query_hdl_per_process(), however,
-   works even before Windows 8. */
-HANDLE
-fhandler_pipe::get_query_hdl_per_system (WCHAR *name,
-					 OBJECT_NAME_INFORMATION *ntfn)
+void
+fhandler_pipe::spawn_worker (int fileno_stdin, int fileno_stdout,
+			     int fileno_stderr)
 {
-  NTSTATUS status;
-  SIZE_T n_handle = 65536;
-  PSYSTEM_HANDLE_INFORMATION shi;
-  do
-    { /* Enumerate handles */
-      SIZE_T nbytes =
-	sizeof (ULONG) + n_handle * sizeof (SYSTEM_HANDLE_TABLE_ENTRY_INFO);
-      shi = (PSYSTEM_HANDLE_INFORMATION) HeapAlloc (GetProcessHeap (),
-						    0, nbytes);
-      if (!shi)
-	return NULL;
-      status = NtQuerySystemInformation (SystemHandleInformation,
-					 shi, nbytes, NULL);
-      if (NT_SUCCESS (status))
-	break;
-      HeapFree (GetProcessHeap (), 0, shi);
-      n_handle *= 2;
-    }
-  while (n_handle < (1L<<23) && status == STATUS_INFO_LENGTH_MISMATCH);
-  if (!NT_SUCCESS (status))
-    return NULL;
+  bool need_send_noncygchld_sig = false;
 
-  for (LONG i = (LONG) shi->NumberOfHandles - 1; i >= 0; i--)
+  /* spawn_worker() is called from spawn.cc only when non-cygwin app
+     is started. Set pipe mode blocking for the non-cygwin process. */
+  int fd;
+  cygheap_fdenum cfd (false);
+  while ((fd = cfd.next ()) >= 0)
+    if (cfd->get_dev () == FH_PIPEW
+	&& (fd == fileno_stdout || fd == fileno_stderr))
+      {
+	fhandler_pipe *pipe = (fhandler_pipe *)(fhandler_base *) cfd;
+	pipe->set_pipe_non_blocking (false);
+
+	/* Setup for query_ndl stuff. Read the comment below. */
+	if (pipe->request_close_query_hdl ())
+	  need_send_noncygchld_sig = true;
+      }
+    else if (cfd->get_dev () == FH_PIPER && fd == fileno_stdin)
+      {
+	fhandler_pipe *pipe = (fhandler_pipe *)(fhandler_base *) cfd;
+	pipe->set_pipe_non_blocking (false);
+      }
+
+  /* If multiple writers including non-cygwin app exist, the non-cygwin
+     app cannot detect pipe closure on the read side when the pipe is
+     created by system account or the pipe creator is running as service.
+     This is because query_hdl which is held in write side also is a read
+     end of the pipe, so the pipe is still alive for the non-cygwin app
+     even after the reader is closed.
+
+     To avoid this problem, let all processes in the same process
+     group close query_hdl using internal signal __SIGNONCYGCHLD when
+     non-cygwin app is started.  */
+  if (need_send_noncygchld_sig)
     {
-      /* Check for the peculiarity of cygwin read pipe */
-      const ULONG access = FILE_READ_DATA | FILE_READ_EA
-	| FILE_WRITE_EA /* marker */
-	| FILE_READ_ATTRIBUTES | FILE_WRITE_ATTRIBUTES
-	| READ_CONTROL | SYNCHRONIZE;
-      if (shi->Handles[i].GrantedAccess != access)
-	continue;
-
-      /* Retrieve handle */
-      HANDLE proc = OpenProcess (PROCESS_DUP_HANDLE, 0,
-				 shi->Handles[i].UniqueProcessId);
-      if (!proc)
-	continue;
-      HANDLE h = (HANDLE)(intptr_t) shi->Handles[i].HandleValue;
-      BOOL res  = DuplicateHandle (proc, h, GetCurrentProcess (), &h,
-				   FILE_READ_DATA, 0, 0);
-      if (!res)
-	goto close_proc;
-
-      /* Check object name */
-      ULONG len;
-      status = NtQueryObject (h, ObjectNameInformation, ntfn, 65536, &len);
-      if (!NT_SUCCESS (status) || !ntfn->Name.Buffer)
-	goto close_handle;
-      ntfn->Name.Buffer[ntfn->Name.Length / sizeof (WCHAR)] = L'\0';
-      if (wcscmp (name, ntfn->Name.Buffer) == 0)
-	{
-	  query_hdl_proc = proc;
-	  query_hdl_value = (HANDLE)(intptr_t) shi->Handles[i].HandleValue;
-	  HeapFree (GetProcessHeap (), 0, shi);
-	  return h;
-	}
-close_handle:
-      CloseHandle (h);
-close_proc:
-      CloseHandle (proc);
+      tty_min dummy_tty;
+      dummy_tty.ntty = (fh_devices) myself->ctty;
+      dummy_tty.pgid = myself->pgid;
+      tty_min *t = cygwin_shared->tty.get_cttyp ();
+      if (!t) /* If tty is not allocated, use dummy_tty instead. */
+	t = &dummy_tty;
+      /* Emit __SIGNONCYGCHLD to let all processes in the
+	 process group close query_hdl. */
+      t->kill_pgrp (__SIGNONCYGCHLD);
     }
-  HeapFree (GetProcessHeap (), 0, shi);
-  return NULL;
 }
